/*
 * $Id: aux-items.h,v 1.13 1999/06/26 11:37:09 ceder Exp $
 * Copyright (C) 1994-1999  Lysator Academic Computer Association.
 *
 * This file is part of the LysKOM server.
 * 
 * LysKOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 1, or (at your option) 
 * any later version.
 * 
 * LysKOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with LysKOM; see the file COPYING.  If not, write to
 * Lysator, c/o ISY, Linkoping University, S-581 83 Linkoping, SWEDEN,
 * or the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
 * MA 02139, USA.
 *
 * Please mail bug reports to bug-lyskom@lysator.liu.se. 
 */
/* Requires kom-types.h */


#define AUX_PREDEFINED_MIN      1
#define AUX_PREDEFINED_MAX      9999
#define AUX_CLIENT_SPECIFIC_MIN 10000
#define AUX_CLIENT_SPECIFIC_MAX 19999
#define AUX_EXPERIMENTAL_MIN    20000
#define AUX_EXPERIMENTAL_MAX    29999

#define AUX_IS_ILLEGAL(x) ((x) < 1)
#define AUX_IS_EXPERIMENTAL(x) ((x) >= AUX_EXPERIMENTAL_MIN && (x) <= AUX_EXPERIMENTAL_MAX)
#define AUX_IS_CLIENT_SPECIFIC(x) ((x) >= AUX_CLIENT_SPECIFIC_MIN && (x) <= AUX_CLIENT_SPECIFIC_MAX)
#define AUX_IS_PREDEFINED(x) (((x) >= AUX_PREDEFINED_MIN && (x) <= AUX_PREDEFINED_MAX) || (x) > AUX_EXPERIMENTAL_MAX)



/* Valid values for text_a etc. */

#define AUX_ITEM_ADD_ON_CREATE 1
#define AUX_ITEM_ADD_ON_MODIFY 2

/* Valid values for action trigger data field */

#define AUX_ITEM_ADD_ACTION    1
#define AUX_ITEM_DELETE_ACTION 2
#define AUX_ITEM_UNDELETE_ACTION 3

typedef struct Aux_item_definition_s Aux_item_definition;

typedef struct Aux_item_trigger_data_s
{
    short           action;
    Object_type     object_type;
    unsigned long   item_index;
    unsigned long   object_no;
    void          * object;
    Aux_item      * item;
} Aux_item_trigger_data;

typedef void (*Aux_item_trigger)(Aux_item_trigger_data *);

typedef enum {
    AUX_VALIDATE_FUNCTION,
    AUX_VALIDATE_REGEXP,
    AUX_VALIDATE_REJECT,
} Aux_item_validation_type;

typedef struct
{
    Aux_item                *item;
    Aux_item_definition     *def;
    Pers_no                  item_creator;
    Pers_no                  object_creator;
    Bool                     owner_check;
    Aux_item_list           *add_to_list;
    unsigned long            start_looking_at;
    Bool                     creating;
    Object_type              object_type;
} Aux_item_validation_data;

typedef Bool (*Aux_item_validation_function)(Aux_item_validation_data *);

typedef struct
{
    Aux_item_validation_type   type;
    union
    {
        struct
        {
            char                        *regexp;
            struct re_pattern_buffer    *cached_re_buf;
        } re;
        struct
        {
            Aux_item_validation_function function;
        } fn;
    } v;
} Aux_item_validator;



typedef struct
{
    const char *name;
    Aux_item_validation_function function;
} Aux_item_validator_mapping;

typedef struct
{
    const char *name;
    Aux_item_trigger function;
} Aux_item_trigger_mapping;

struct Aux_item_definition_s {
    char                   *  name;
    unsigned long             tag;
    Aux_item_flags            clear_flags;
    Aux_item_flags            set_flags;
    Bool                      disabled;
    Bool                      author_only;
    Bool                      supervisor_only;
    Bool                      system_only;
    Bool                      one_per_person;
    Bool                      may_not_delete;
    unsigned long             inherit_limit;
    Bool                      texts;
    short                     text_a; /* When can we add items of this type */
    Bool                      confs;
    short                     conf_a; /* When can we add items of this type */
    Bool                      letterboxes;
    Bool                      system;
    unsigned long             num_validators;
    Aux_item_validator      * validators;
    unsigned long             num_delete_triggers;
    Aux_item_trigger        * delete_triggers;
    unsigned long             num_add_triggers;
    Aux_item_trigger        * add_triggers;
    unsigned long             num_undelete_triggers;
    Aux_item_trigger        * undelete_triggers;
    struct Aux_item_definition_s *next;
};


extern Aux_item_definition empty_aux_item_definition;

/* Inerit items from parent to target. counter is a pointer to
 * highest_aux_item or equivalent. target_creator is the creator
 * of the target object (author for texts, person for persons and
 * conference for conferences */

void aux_inherit_items(Aux_item_list *target, Aux_item_list *parent,
                       unsigned long *counter, Pers_no target_creator,
                       Bool creating, Object_type object_type,
                       unsigned long object_no, void *object);



/* Check for permission to add item to some object. def is NULL or the
 * item definition. item_creator is the person creating the item.
 * object_creator is the person who created the object we're adding
 * the item to. owner_check is FALSE only if we don't want to check
 * author and supervisor related definition flags (when creating a
 * person or conference, for example.) add_to_list is the list of
 * items already on the object or NULL, and start_looking_at is the
 * first position in add_to_list that we should look at.
 *
 * If you think you need to call this, and you're not hacking functions
 * in aux-items.c, you should rethink your strategy. */

Bool aux_item_add_perm(Aux_item *item, Aux_item_definition *def,
                       Pers_no item_creator, Pers_no object_creator,
                       Bool owner_check, Aux_item_list *add_to_list,
                       unsigned long start_looking_at, Bool creating,
                       Object_type object_type);




/* Prepare the items in list for addition to an object. creator is the
 * person who will add the items. The items in list will be modified.
 * You need to call this before sending the list to any of the other
 * functions. */

void prepare_aux_item_list(Aux_item_list *list, Pers_no creator);
void prepare_aux_item(Aux_item *item, Pers_no creator,
                      Aux_item_definition *def);



/* Find the aux_item_definition for item. Returns a pointer to the
 * definition or NULL if the item's tag is invalid. */

Aux_item_definition *find_aux_item_definition(Aux_item *item);


/* Get the function pointer for a named trigger */

Aux_item_trigger aux_item_find_trigger(char *trigger_name);

/* Get the function pointer for a named validation function */

Aux_item_validation_function aux_item_find_validator(char *validator_name);

/* Get a pointer to the item in list with aux-no aux_no. Returns
 * NULL if there is no such item */

long find_aux_item_index(Aux_item_list *list, unsigned long aux_no);
Aux_item *find_aux_item(Aux_item_list *list, unsigned long aux_no);



/* Filter the items in original for sending to viewer. Result is a
 * pointer to an empty aux_item_list, in which copies of the items in
 * original will be places. All memory allocated in this call is
 * allocaed with tmp_alloc */

void filter_aux_item_list(Aux_item_list *original,
                          Aux_item_list *result,
                          Connection *conn);


/* delete items_to_delete from list_to_delete_from
 * list_to_delete_from and its elements may be modified */

void delete_aux_item_list(Number_list *items_to_delete,
                          Aux_item_list *list_to_delete_from,
                          Object_type object_type,
                          unsigned long object_no,
                          void *object);



/* undelete items in items_to_undelete from list_to_undelete_from
 * list_to_delete_from and its elements may be modified */

void undelete_aux_item_list(Number_list *items_to_undelete,
                            Aux_item_list *list_to_undelete_from,
                            Object_type object_type,
                            unsigned long object_no,
                            void *object);


void
commit_aux_item_list(Aux_item_list *list_to_commit);


/* Check if we may delete items_to_delete from list_to_delete_from
 * Elements in items_to_delete may be modified. Sets kom_errno,
 * err_stat and returns FAILURE if deletion of any single item is
 * not allowed */

Success check_delete_aux_item_list(Number_list *items_to_delete,
                                   Aux_item_list *list_to_delete_from);



/* Att item_list to the aux_item_list of text_s. item_creator is the
 * person who is adding the items. Does not check for permission first */
  
void text_stat_add_aux_item_list(Text_stat *text_s,
                                 Text_no text_no,
                                 Aux_item_list *item_list,
                                 Pers_no item_creator);



/* Check for permission to add items in list to aux_item_list
 * of text_s. item_creator is the person adding the items. Sets
 * kom_errno and err_stat and returns FAILURE if permission is
 * not granted. You MUST call perpare_aux_item_list on the list
 * before calling this! */

Success text_stat_check_add_aux_item_list(Text_stat *text_s,
                                          Aux_item_list *list,
                                          Pers_no item_creator,
                                          Bool creating);



Success conf_stat_check_add_aux_item(Conference          *conf,
                                     Conf_no        conf_no,
                                     Aux_item            *item,
                                     Pers_no              creator,
                                     Aux_item_definition *def,
                                     Bool creating);

Success
conf_stat_check_add_aux_item_list(Conference    *conf,
                                  Conf_no        conf_no,
                                  Aux_item_list *list,
                                  Pers_no        creator,
                                  Bool creating);
void
conf_stat_add_aux_item(Conference          *conf,
                       Conf_no              conf_no,
                       Aux_item            *item,
                       Aux_item_definition *def);
void
conf_stat_add_aux_item_list(Conference    *conf,
                            Conf_no        conf_no,
                            Aux_item_list *list,
                            Pers_no        creator);

void
system_add_aux_item_list(Info    *info,
                         Aux_item_list *item_list,
                         Pers_no        item_creator);

Success system_check_add_aux_item_list(Info *info,
                                       Aux_item_list *list,
                                       Pers_no creator);

void initialize_aux_items(char *);
void free_aux_item_definitions(void);

void aux_item_definition_add(Aux_item_definition *);

/* Parser function from aux-item-def.y. */
extern void parse_aux_item_definitions(char *file);


